# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from typing import Any, Optional, Union, TYPE_CHECKING, Protocol, BinaryIO, Literal
from ._typing_compat import (
    Dict, List, Tuple, Callable, Awaitable,
    Iterator, AsyncIterator, Mapping, TypeAlias
)

if TYPE_CHECKING:  # pragma: no cover
    from ._session import Session, Response
    from ._async_session import AsyncSession, AsyncResponse
    from ._client import Client
    from ._async_client import AsyncClient

__all__ = [
    "JSON",
    "AnyClient",
    "AnyResponse",
    "AsyncConsumeCallback",
    "AsyncFileLike",
    "AsyncFileOrPath",
    "AsyncFileOrPathDestination",
    "AsyncOpenFileCallback",
    "AsyncPayload",
    "AsyncSessionFactory",
    "AsyncSessionName",
    "BinaryAsyncFileLike",
    "ConsumeCallback",
    "FileOpenMode",
    "FileOrPath",
    "FileOrPathDestination",
    "HTTPMethod",
    "Headers",
    "OpenFileCallback",
    "OperationStatus",
    "Payload",
    "SessionFactory",
    "SessionName",
    "TimeoutParameter"
]

#: JSON data (parsed)
JSON: TypeAlias = Union[Dict, List, str, int, float, None]

#: Request timeout (in seconds). Can be a single number, None or a tuple.
#: If the timeout is specified as a tuple, then the first value is the
#: connect timeout, and the second value is the read timeout.
#: Otherwise, both connect and read timeouts are set to the same value.
#: A value of None means no timeout.
#: If the timeout's value is :code:`...`, the default timeout is used
#: (either :any:`settings.DEFAULT_TIMEOUT` or :any:`settings.DEFAULT_UPLOAD_TIMEOUT`)
TimeoutParameter: TypeAlias = Optional[
    Union[
        float,
        Tuple[Optional[float], Optional[float]]
    ]
]

#: Type used for passing HTTP request headers
Headers: TypeAlias = Mapping[str, str]

#: Request payload - data to be uploaded
Payload: TypeAlias = Union[bytes, Iterator[bytes]]

#: Callback function that is invoked to consume the streamed HTTP response body
ConsumeCallback: TypeAlias = Callable[[bytes], None]

#: Request payload - data to be uploaded (async variant)
AsyncPayload: TypeAlias = Union[bytes, Iterator[bytes], AsyncIterator[bytes]]

#: Callback function (may be asynchronous) that is invoked to consume the
#: streamed HTTP response body
AsyncConsumeCallback: TypeAlias = Union[Callable[[bytes], None], Callable[[bytes], Awaitable[None]]]

#: :any:`Response` or :any:`AsyncResponse`
AnyResponse: TypeAlias = Union["Response", "AsyncResponse"]

#: :any:`Session` or :any:`AsyncSession`
AnySession: TypeAlias = Union["Session", "AsyncSession"]

#: :any:`Client` or :any:`AsyncClient`
AnyClient: TypeAlias = Union["Client", "AsyncClient"]


class AsyncFileLike(Protocol):
    """
        This protocol describes the bare minimum set of required methods for
        an async file-like object (open in either binary or unicode mode).
    """

    async def read(self, size: int = ..., /) -> Union[str, bytes]:
        """
            Reads `size` bytes or characters.

            :param size: `int`, number of bytes/characters to read from the file
            :returns: data that was read from the file
        """
        ...

    async def write(self, buffer: Any, /) -> int:
        """
            Writes data (contained in `buffer`).

            :param buffer: data to be written
            :returns: the number of written bytes/characters
        """
        ...

    async def seek(self, pos: int, whence: int = ..., /) -> int:
        """
            Performs a seek operation on a file.

            :param pos: `int`, position to seek to
            :param whence: `int`, 0 (seek absolute position), 1 (seek relative
                            to current position) or 2 (seek to file's end)

            :returns: `int`, absolute position within the file after the seek operation
        """
        ...

    async def tell(self) -> int:
        """
            Returns current position within the file.

            :returns: `int`, current position within the file
        """
        ...


class BinaryAsyncFileLike(Protocol):
    """
        This protocol describes the bare minimum set of required methods for
        an async file-like object open in binary mode.
    """

    async def read(self, size: int = ..., /) -> bytes:
        """
            Reads `size` bytes.

            :param size: `int`, number of bytes/characters to read from the file
            :returns: data that was read from the file
        """
        ...

    async def write(self, buffer: Any, /) -> int:
        """
            Writes data (contained in `buffer`).

            :param buffer: data to be written
            :returns: the number of written bytes
        """
        ...

    async def seek(self, pos: int, whence: int = ..., /) -> int:
        """
            Performs a seek operation on a file.

            :param pos: `int`, position to seek to
            :param whence: `int`, 0 (seek absolute position), 1 (seek relative
                            to current position) or 2 (seek to file's end)

            :returns: `int`, absolute position within the file after the seek operation
        """
        ...

    async def tell(self) -> int:
        """
            Returns current position within the file.

            :returns: `int`, current position within the file
        """
        ...


#: This is used to specify a source file to upload
FileOrPath: TypeAlias = Union[
    str,
    bytes,
    BinaryIO,
    Callable[[], Iterator[bytes]]
]

#: This is used to specify a destination file to download into
FileOrPathDestination: TypeAlias = Union[
    str,
    bytes,
    BinaryIO,
]

#: This is used to specify a source file to upload (async variant)
AsyncFileOrPath: TypeAlias = Union[
    str,
    bytes,
    BinaryIO,
    AsyncFileLike,
    Callable[[], AsyncIterator[bytes]]
]

#: This is used to specify a destination file to download into (async variant)
AsyncFileOrPathDestination: TypeAlias = Union[
    str,
    bytes,
    BinaryIO,
    BinaryAsyncFileLike
]

#: Function that returns an instance of :any:`Session`
SessionFactory: TypeAlias = Callable[[], "Session"]

#: Function that returns an instance of :any:`AsyncSession`
AsyncSessionFactory: TypeAlias = Callable[[], "AsyncSession"]

#: File mode for :any:`OpenFileCallback` and :any:`AsyncOpenFileCallback`
FileOpenMode: TypeAlias = Union[Literal["rb"], Literal["wb"]]

#: Function that is used for opening local files (like :any:`open`)
OpenFileCallback: TypeAlias = Callable[[Union[str, bytes], FileOpenMode], BinaryIO]

#: Function that is used for opening local files (async variant)
AsyncOpenFileCallback: TypeAlias = Union[
    Callable[[Union[str, bytes], FileOpenMode], Awaitable[BinaryAsyncFileLike]],
    Callable[[Union[str, bytes], FileOpenMode], Awaitable[BinaryIO]],
]

#: HTTP request method
HTTPMethod: TypeAlias = Union[
    Literal["GET"],
    Literal["POST"],
    Literal["PUT"],
    Literal["PATCH"],
    Literal["DELETE"],
    Literal["OPTIONS"],
    Literal["HEAD"],
    Literal["CONNECT"],
    Literal["TRACE"],
]

#: Valid session name (see :doc:`/api_reference/sessions`)
SessionName: TypeAlias = Union[Literal["httpx"], Literal["pycurl"], Literal["requests"]]

#: Valid asynchronous session name (see :doc:`/api_reference/sessions`)
AsyncSessionName: TypeAlias = Union[Literal["aiohttp"], Literal["httpx"]]

#: Yandex.Disk's asynchronous operation status
OperationStatus: TypeAlias = Union[Literal["in-progress"], Literal["success"], Literal["failed"]]



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

import logging
from .types import TimeoutParameter

__all__ = [
    "BASE_API_URL",
    "BASE_OAUTH_API_URL",
    "DEFAULT_N_RETRIES",
    "DEFAULT_RETRY_INTERVAL",
    "DEFAULT_TIMEOUT",
    "DEFAULT_UPLOAD_RETRY_INTERVAL",
    "DEFAULT_UPLOAD_TIMEOUT",
    "logger"
]

#: `tuple` of 2 numbers (`int` or `float`), default timeout for requests.
#: First number is the connect timeout, the second one is the read timeout.
DEFAULT_TIMEOUT: TimeoutParameter = (10.0, 15.0)

#: `int`, default number of retries
DEFAULT_N_RETRIES: int = 3

#: `float`, default retry interval
DEFAULT_RETRY_INTERVAL: float = 0.0

#: Analogous to :any:`settings.DEFAULT_TIMEOUT` but for
#: :any:`Client.upload()`/:any:`AsyncClient.upload()` function
DEFAULT_UPLOAD_TIMEOUT: TimeoutParameter = DEFAULT_TIMEOUT

#: Analogous to :any:`settings.DEFAULT_RETRY_INTERVAL` but for
#: :any:`Client.upload()`/:any:`AsyncClient.upload()` function
DEFAULT_UPLOAD_RETRY_INTERVAL: float = 0.0

#: Base URL for Yandex.Disk's REST API.
#: Can be overriden for testing and other purposes
BASE_API_URL: str = "https://cloud-api.yandex.net"

#: Base URL for Yandex.Disk's OAuth API.
#: Can be overriden for testing and other purposes
BASE_OAUTH_API_URL: str = "https://oauth.yandex.ru"

#: Logger for the library. Logs include information about requests to the API
#: and automatic retry attempts.
logger = logging.getLogger("yadisk")



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from pathlib import PurePosixPath
import posixpath
import time
from urllib.parse import urlencode

from ._api import *

from .exceptions import (
    AsyncOperationFailedError, AsyncOperationPollingTimeoutError, ParentNotFoundError,
    PathNotFoundError, RetriableYaDiskError, UnauthorizedError,
    OperationNotFoundError, InvalidResponseError, WrongResourceTypeError
)

from .utils import auto_retry, CaseInsensitiveDict
from .objects import (
    SyncResourceLinkObject, SyncPublicResourceLinkObject,
    SyncTrashResourceObject, SyncFilesResourceListObject, SyncResourceObject,
    SyncLastUploadedResourceListObject, SyncOperationLinkObject,
    SyncPublicResourceObject, SyncPublicResourcesListObject, DiskInfoObject,
    TokenObject, TokenRevokeStatusObject, DeviceCodeObject, ResourceUploadLinkObject
)

from ._session import Session
from ._import_session import import_session

from . import settings

from typing import Any, Optional, Union, Literal
from ._typing_compat import Callable, Generator, Dict, List, Type
from .types import (
    OpenFileCallback, FileOrPath, FileOrPathDestination, OperationStatus,
    SessionFactory, SessionName
)

from ._client_common import (
    _add_spoof_user_agent_header, _apply_default_args, _filter_request_kwargs,
    _read_file_as_generator, _set_authorization_header,
    _add_authorization_header, _validate_listdir_response,
    _validate_link_response, _validate_get_type_response
)

from ._common import remove_path_schema

__all__ = ["Client"]

ResourceType = Union[
    "SyncResourceObject",
    "SyncPublicResourceObject",
    "SyncTrashResourceObject"
]


def _exists(
    get_meta_function: Callable[..., ResourceType],
    /,
    *args,
    **kwargs
) -> bool:
    try:
        # We want to query the bare minimum number of fields, that's what
        # the fields parameter is for
        get_meta_function(*args, fields=["type"], **kwargs)

        return True
    except PathNotFoundError:
        return False


def _get_type(
    get_meta_function: Callable[..., ResourceType],
    /,
    *args,
    **kwargs
) -> str:
    return get_meta_function(
        *args,
        _then=_validate_get_type_response,
        fields=["type"],
        **kwargs
    ).type  # type: ignore[return-value]


def _listdir(
    get_meta_function: Callable[..., ResourceType],
    path: str,
    /,
    *,
    max_items: Optional[int] = None,
    **kwargs
) -> Generator[Any, None, None]:
    if kwargs.get("limit") is None:
        kwargs["limit"] = 500

    if kwargs.get("fields") is None:
        kwargs["fields"] = []

    kwargs["fields"] = ["embedded.items.%s" % (k,) for k in kwargs["fields"]]

    # Fields that are absolutely necessary
    NECESSARY_FIELDS = ["type",
                        "embedded",
                        "embedded.offset",
                        "embedded.limit",
                        "embedded.total",
                        "embedded.items"]

    kwargs["fields"].extend(NECESSARY_FIELDS)

    remaining_items = max_items

    if remaining_items is not None:
        # Do not query more items than necessary
        kwargs["limit"] = min(remaining_items, kwargs["limit"])

    result = get_meta_function(path, _then=_validate_listdir_response, **kwargs)

    if result.type == "file":
        raise WrongResourceTypeError("%r is a file" % (path,))

    yield from result.embedded.items[:remaining_items]  # type: ignore[union-attr,index]

    limit: int = result.embedded.limit  # type: ignore[assignment,union-attr]
    offset: int = result.embedded.offset  # type: ignore[assignment,union-attr]
    total: int = result.embedded.total  # type: ignore[assignment,union-attr]

    while offset + limit < total:
        if remaining_items is not None:
            remaining_items -= len(result.embedded.items)  # type: ignore[union-attr,arg-type]

            if remaining_items <= 0:
                break

            # Do not query more items than necessary
            kwargs["limit"] = min(remaining_items, kwargs["limit"])
        else:
            remaining_items = None

        offset += limit
        kwargs["offset"] = offset
        result = get_meta_function(path, _then=_validate_listdir_response, **kwargs)

        if result.type == "file":
            raise WrongResourceTypeError("%r is a file" % (path,))

        yield from result.embedded.items[:remaining_items]  # type: ignore[union-attr,index]

        limit = result.embedded.limit  # type: ignore[assignment,union-attr]
        total = result.embedded.total  # type: ignore[assignment,union-attr]


class Client:
    """
        Implements access to Yandex.Disk REST API (provides synchronous API).

        HTTP client implementation can be specified using the :code:`session`
        parameter. :any:`RequestsSession` is used by default. For other options,
        see :doc:`/api_reference/sessions`.

        Almost all methods of :any:`Client` (the ones that accept `**kwargs`)
        accept some additional arguments:

        * **n_retries** - `int`, maximum number of retries for a request
        * **retry_interval** - `float`, delay between retries (in seconds)
        * **headers** - `dict` or `None`, additional request headers
        * **timeout** - `tuple` (:code:`(<connect timeout>, <read timeout>)`) or
          `float` (specifies both connect and read timeout), request timeout
          (in seconds)

        Additional parameters, specific to a given HTTP client library can also
        be passed, see documentation for specific :any:`Session` subclasses
        (:doc:`/api_reference/sessions`).

        :param id: application ID
        :param secret: application secret password
        :param token: application token
        :param default_args: `dict` or `None`, default arguments for methods.
                             Can be used to set the default timeout, headers, etc.
        :param session: `None`, `str` or an instance of :any:`Session`.
                        If :code:`session` is a string, the appropriate session
                        class will be imported, it must be one of the
                        following values:

                          * :code:`"httpx"` - :any:`HTTPXSession`
                          * :code:`"pycurl"` - :any:`PycURLSession`
                          * :code:`"requests"` - :any:`RequestsSession`

        :param open_file: `None` or a function that opens a file for reading or
                          writing (:code:`open()` by default)
        :param session_factory: kept for compatibility, callable that returns an
                                instance of :any:`Session`

        :ivar id: `str`, application ID
        :ivar secret: `str`, application secret password
        :ivar token: `str`, application token
        :ivar default_args: `dict`, default arguments for methods. Can be used to
                            set the default timeout, headers, etc.
        :ivar session: current session (:any:`Session` instance)
        :ivar open_file: function that opens a file for reading or writing
                         (:code:`open()` by default)

        The following exceptions may be raised by most API requests:

        :raises RequestError: HTTP client raised an exception while making a request
        :raises BadRequestError: server returned HTTP code 400
        :raises FieldValidationError: request contains fields with invalid data
        :raises UnauthorizedError: server returned HTTP code 401
        :raises ForbiddenError: server returned HTTP code 403
        :raises NotAcceptableError: server returned HTTP code 406
        :raises ConflictError: server returned HTTP code 409
        :raises PayloadTooLargeError: server returned code 413
        :raises UnsupportedMediaError: server returned HTTP code 415
        :raises LockedError: server returned HTTP code 423
        :raises TooManyRequestsError: server returned HTTP code 429
        :raises InternalServerError: server returned HTTP code 500
        :raises BadGatewayError: server returned HTTP code 502
        :raises UnavailableError: server returned HTTP code 503
        :raises GatewayTimeoutError: server returned HTTP code 504
        :raises InsufficientStorageError: server returned HTTP code 509
        :raises UnknownYaDiskError: other unknown error
    """

    id: str
    secret: str
    token: str
    default_args: Dict[str, Any]
    session: Session
    open_file: OpenFileCallback

    synchronous = True

    def __init__(self,
                 id:     str = "",
                 secret: str = "",
                 token:  str = "",
                 *,
                 default_args:    Optional[Dict[str, Any]] = None,
                 session:         Optional[Union[Session, SessionName]] = None,
                 open_file:       Optional[OpenFileCallback] = None,
                 session_factory: Optional[SessionFactory] = None) -> None:
        self.id = id
        self.secret = secret
        self.token = ""

        self.default_args = {} if default_args is None else default_args

        if open_file is None:
            open_file = open

        self.open_file = open_file

        if session is None:
            if session_factory is not None:
                session = session_factory()
            else:
                try:
                    session = import_session("requests")()
                except ModuleNotFoundError as e:
                    if e.name == "requests":
                        raise ModuleNotFoundError(
                            "requests is not installed. Either install requests or provide a custom session",
                            name=e.name,
                            path=e.path) from e
                    else:
                        raise
        elif isinstance(session, str):
            session = import_session(session)()

        self.session = session
        self.token = token

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs) -> None:
        self.close()

    def close(self) -> None:
        """
            Closes the session.
            Do not call this method while there are other active threads using this object.

            This method can also be called implicitly by using the `with`
            statement.
        """

        self.session.close()

    def _maybe_wait(
        self,
        request_class: Type[APIRequest],
        /,
        *args,
        wait: bool = True,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        **kwargs
    ) -> Any:
        request = request_class(self.session, *args, **kwargs)

        if wait:
            args_to_filter = (
                "permanently", "md5", "overwrite", "force_async", "fields"
            )

            for arg in args_to_filter:
                kwargs.pop(arg, None)

            def then(response: Optional[SyncOperationLinkObject]) -> Optional[SyncOperationLinkObject]:
                if not isinstance(response, SyncOperationLinkObject):
                    return response

                try:
                    response.wait(
                        poll_interval=poll_interval,
                        poll_timeout=poll_timeout,
                        **kwargs
                    )
                except RetriableYaDiskError as e:
                    # We want to trigger a full retry (including the operation iteslf)
                    # only if the asynchronous operation failed
                    if not isinstance(e, AsyncOperationFailedError):
                        e.disable_retry = True
                    else:
                        settings.logger.info("asynchronous operation failed, attempting to restart it")

                    raise e from None

                return response

            return request.send(yadisk=self, then=then)
        else:
            return request.send(yadisk=self)

    def get_auth_url(
        self,
        type:                  Union[Literal["code"], Literal["token"]],
        device_id:             Optional[str] = None,
        device_name:           Optional[str] = None,
        redirect_uri:          Optional[str] = None,
        login_hint:            Optional[str] = None,
        scope:                 Optional[str] = None,
        optional_scope:        Optional[str] = None,
        force_confirm:         bool = True,
        state:                 Optional[str] = None,
        code_challenge:        Optional[str] = None,
        code_challenge_method: Optional[Union[Literal["plain"], Literal["S256"]]] = None,
        display:               None = None
    ) -> str:
        """
            Get authentication URL for the user to go to.
            This method doesn't send any HTTP requests and merely constructs the URL.

            :param type: response type ("code" to get the confirmation code or "token" to get the token automatically)
            :param device_id: unique device ID, must be between 6 and 50 characters
            :param device_name: device name, should not be longer than 100 characters
            :param redirect_uri: the URL to redirect the user to after they allow access to the app,
                                 by default, the first redirect URI specified in the app settings
                                 is used
            :param display: doesn't do anything, kept for compatibility
            :param login_hint: username or email for the account the token is being requested for
            :param scope: `str`, list of permissions for the application
            :param optional_scope: `str`, list of optional permissions for the application
            :param force_confirm: if True, user will be required to confirm access to the account
                                  even if the user has already granted access for the application
            :param state: The state string, which Yandex.OAuth returns without any changes (<= 1024 characters)
            :param code_challenge: string derived from the generated :code:`code_verifier` value
                                   using one of the two possible transformations (plain or S256)
            :param code_challenge_method: specifies what function was used to transform
                                          the :code:`code_verifier` value to :code:`code_challenge`,
                                          allowed values are :code:`"plain"` and :code:`"S256"` (recommended).
                                          If :code:`"S256"` is used, :code:`code_challenge` must be produced
                                          by hashing the :code:`code_verifier` value and encoding it to base64

            :raises ValueError: invalid arguments were passed

            :returns: authentication URL
        """

        if type not in ("code", "token"):
            raise ValueError("type must be either 'code' or 'token'")

        if code_challenge_method not in (None, "plain", "S256"):
            raise ValueError("code_challenge_method must be either 'plain' or 'S256'")

        params = {"response_type": type,
                  "client_id":     self.id,
                  "force_confirm": "yes" if force_confirm else "no"}

        if device_id is not None:
            params["device_id"] = device_id

        if device_name is not None:
            params["device_name"] = device_name

        if redirect_uri is not None:
            params["redirect_uri"] = redirect_uri

        if login_hint is not None:
            params["login_hint"] = login_hint

        if scope is not None:
            params["scope"] = " ".join(scope)

        if optional_scope is not None:
            params["optional_scope"] = " ".join(optional_scope)

        if state is not None:
            params["state"] = state

        if code_challenge is not None:
            params["code_challenge"] = code_challenge

        if code_challenge_method is not None:
            params["code_challenge_method"] = code_challenge_method

        return "https://oauth.yandex.ru/authorize?" + urlencode(params)

    def get_code_url(
        self,
        device_id:             Optional[str] = None,
        device_name:           Optional[str] = None,
        redirect_uri:          Optional[str] = None,
        login_hint:            Optional[str] = None,
        scope:                 Optional[str] = None,
        optional_scope:        Optional[str] = None,
        force_confirm:         bool = True,
        state:                 Optional[str] = None,
        code_challenge:        Optional[str] = None,
        code_challenge_method: Optional[Union[Literal["plain"], Literal["S256"]]] = None,
        display:               None = None
    ) -> str:
        """
            Get the URL for the user to get the confirmation code.
            The confirmation code can later be used to get the token.
            This method doesn't send any HTTP requests and merely constructs the URL.

            :param device_id: unique device ID, must be between 6 and 50 characters
            :param device_name: device name, should not be longer than 100 characters
            :param redirect_uri: the URL to redirect the user to after they allow access to the app,
                                 by default, the first redirect URI specified in the app settings
                                 is used
            :param display: doesn't do anything, kept for compatibility
            :param login_hint: username or email for the account the token is being requested for
            :param scope: `str`, list of permissions for the application
            :param optional_scope: `str`, list of optional permissions for the application
            :param force_confirm: if True, user will be required to confirm access to the account
                                  even if the user has already granted access for the application
            :param state: The state string, which Yandex.OAuth returns without any changes (<= 1024 characters)
            :param code_challenge: string derived from the generated :code:`code_verifier` value
                                   using one of the two possible transformations (plain or S256)
            :param code_challenge_method: specifies what function was used to transform
                                          the :code:`code_verifier` value to :code:`code_challenge`,
                                          allowed values are :code:`"plain"` and :code:`"S256"` (recommended).
                                          If :code:`"S256"` is used, :code:`code_challenge` must be produced
                                          by hashing the :code:`code_verifier` value and encoding it to base64

            :raises ValueError: invalid arguments were passed

            :returns: authentication URL
        """

        return self.get_auth_url(
            "code",
            device_id=device_id,
            device_name=device_name,
            redirect_uri=redirect_uri,
            display=display,
            login_hint=login_hint,
            scope=scope,
            optional_scope=optional_scope,
            force_confirm=force_confirm,
            state=state,
            code_challenge=code_challenge,
            code_challenge_method=code_challenge_method
        )

    def get_device_code(self, **kwargs) -> "DeviceCodeObject":
        """
            This request is used for authorization using the Yandex OAuth page.
            In this case the user must enter the verification code (:code:`user_code`)
            in the browser on the Yandex OAuth page.
            After the user has entered the code on the OAuth page, the application
            can exchange the :code:`device_code` for the token using the :any:`Client.get_token_from_device_code()`.

            :param device_id: unique device ID (between 6 and 50 characters)
            :param device_name: device name, should not be longer than 100 characters
            :param scope: `str`, list of permissions for the application
            :param optional_scope: `str`, list of optional permissions for the application
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises InvalidClientError: invalid client ID
            :raises BadRequestError: invalid request parameters

            :returns: :any:`DeviceCodeObject` containing :code:`user_code` and :code:`device_code`
        """

        _apply_default_args(kwargs, self.default_args)
        _set_authorization_header(kwargs, "")

        return GetDeviceCodeRequest(self.session, self.id, **kwargs).send(yadisk=self)

    def get_token(self, code: str, /, **kwargs) -> "TokenObject":
        """
            Get a new token.

            :param code: confirmation code
            :param device_id: unique device ID (between 6 and 50 characters)
            :param device_name: device name, should not be longer than 100 characters
            :param code_verifier: `str`, verifier code, used with the PKCE authorization flow
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises BadVerificationCodeError: confirmation code has invalid format
            :raises InvalidGrantError: invalid or expired confirmation code
            :raises InvalidClientError: invalid client ID or client secret
            :raises BadRequestError: invalid request parameters

            :returns: :any:`TokenObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _set_authorization_header(kwargs, "")

        return GetTokenRequest(
            self.session,
            "authorization_code",
            client_id=self.id,
            code=code,
            client_secret=self.secret,
            **kwargs
        ).send(yadisk=self)

    def get_token_from_device_code(self, device_code: str, /, **kwargs) -> "TokenObject":
        """
            Get a new token from a device code, previously obtained with :any:`Client.get_device_code()`.

            :param device_code: device code, obtained from :any:`Client.get_device_code()`
            :param device_id: unique device ID (between 6 and 50 characters)
            :param device_name: device name, should not be longer than 100 characters
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises AuthorizationPendingError: user has not authorized the application yet
            :raises BadVerificationCodeError: :code:`device_code` has invalid format
            :raises InvalidGrantError: invalid or expired :code:`device_code`
            :raises InvalidClientError: invalid client ID or client secret
            :raises BadRequestError: invalid request parameters

            :returns: :any:`TokenObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _set_authorization_header(kwargs, "")

        return GetTokenRequest(
            self.session,
            "device_code",
            client_id=self.id,
            code=device_code,
            client_secret=self.secret,
            **kwargs
        ).send(yadisk=self)

    def refresh_token(self, refresh_token: str, /, **kwargs) -> "TokenObject":
        """
            Refresh an existing token.

            :param refresh_token: the refresh token that was received with the token
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises InvalidGrantError: invalid or expired refresh token or it
                                       doesn't belong to this application
            :raises InvalidClientError: invalid client ID or client secret
            :raises BadRequestError: invalid request parameters

            :returns: :any:`TokenObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _set_authorization_header(kwargs, "")

        return RefreshTokenRequest(
            self.session,
            refresh_token,
            self.id,
            self.secret,
            **kwargs
        ).send(yadisk=self)

    def revoke_token(
        self,
        token: Optional[str] = None,
        /,
        **kwargs
    ) -> "TokenRevokeStatusObject":
        """
            Revoke the token.

            :param token: token to revoke
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises InvalidGrantError: specified token doesn't belong to this application
            :raises InvalidClientError: invalid client ID or client secret
            :raises UnsupportedTokenTypeError: token could not be revoked because
                                               it doesn't have a :code:`device_id`
            :raises BadRequestError: invalid request parameters

            :returns: :any:`TokenRevokeStatusObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _set_authorization_header(kwargs, "")

        if token is None:
            token = self.token

        return RevokeTokenRequest(
            self.session, token, self.id, self.secret, **kwargs
        ).send(yadisk=self)

    def check_token(self, token: Optional[str] = None, /, **kwargs) -> bool:
        """
            Check whether the token is valid.

            :param token: token to check, equivalent to `self.token` if `None`
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :returns: `bool`
        """

        # Any ID will do, doesn't matter whether it exists or not
        fake_operation_id = "0000"

        if token is None:
            token = self.token

        if not token:
            return False

        _set_authorization_header(kwargs, token)

        try:
            # get_operation_status() doesn't require any permissions, unlike most other requests
            self.get_operation_status(fake_operation_id, **kwargs)
            return True
        except OperationNotFoundError:
            return True
        except UnauthorizedError:
            return False

    def get_disk_info(self, **kwargs) -> "DiskInfoObject":
        """
            Get disk information.

            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: :any:`DiskInfoObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return DiskInfoRequest(self.session, **kwargs).send(yadisk=self)

    def get_meta(self, path: str, /, **kwargs) -> "SyncResourceObject":
        """
            Get meta information about a file/directory.

            :param path: path to the resource
            :param limit: number of children resources to be included in the response
            :param offset: number of children resources to be skipped in the response
            :param preview_size: size of the file preview
            :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
            :param sort: `str`, field to be used as a key to sort children resources
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: :any:`SyncResourceObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        # This is for internal error handling
        _then = kwargs.pop("_then", None)

        return GetMetaRequest(self.session, path, **kwargs).send(yadisk=self, then=_then)

    def exists(self, path: str, /, **kwargs) -> bool:
        """
            Check whether `path` exists.

            :param path: path to the resource
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: `bool`
        """

        return _exists(self.get_meta, path, **kwargs)

    def get_type(self, path: str, /, **kwargs) -> str:
        """
            Get resource type.

            :param path: path to the resource
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: "file" or "dir"
        """

        return _get_type(self.get_meta, path, **kwargs)

    def is_file(self, path: str, /, **kwargs) -> bool:
        """
            Check whether `path` is a file.

            :param path: path to the resource
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: `True` if `path` is a file, `False` otherwise (even if it doesn't exist)
        """

        try:
            return self.get_type(path, **kwargs) == "file"
        except PathNotFoundError:
            return False

    def is_dir(self, path: str, /, **kwargs) -> bool:
        """
            Check whether `path` is a directory.

            :param path: path to the resource
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: `True` if `path` is a directory, `False` otherwise (even if it doesn't exist)
        """

        try:
            return self.get_type(path, **kwargs) == "dir"
        except PathNotFoundError:
            return False

    def listdir(self, path: str, /, **kwargs) -> Generator["SyncResourceObject", None, None]:
        """
            Get contents of `path`.

            :param path: path to the directory
            :param max_items: `int` or `None`, maximum number of returned items (`None` means unlimited)
            :param limit: number of children resources to be included in the response
            :param offset: number of children resources to be skipped in the response
            :param preview_size: size of the file preview
            :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises WrongResourceTypeError: resource is not a directory

            :returns: generator of :any:`ResourceObject`
        """

        return _listdir(self.get_meta, path, **kwargs)

    def get_upload_link(
        self,
        path: str,
        /,
        spoof_user_agent: bool = True,
        **kwargs
    ) -> str:
        """
            Get a link to upload the file using the PUT request.

            :param path: destination path
            :param overwrite: `bool`, determines whether to overwrite the destination
            :param spoof_user_agent: `bool`, if `True` (default), the `User-Agent` header
                will be set to a special value, which should allow bypassing of
                Yandex.Disk's upload speed limit
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ParentNotFoundError: parent directory doesn't exist
            :raises PathExistsError: destination path already exists
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request
            :raises InsufficientStorageError: cannot upload file due to lack of storage space
            :raises UploadTrafficLimitExceededError: upload limit has been exceeded

            :returns: `str`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        # This is used to bypass Yandex.Disk's upload speed limit for some file types
        if spoof_user_agent:
            _add_spoof_user_agent_header(kwargs)

        return GetUploadLinkRequest(
            self.session, path, fields=["href"], **kwargs
        ).send(yadisk=self, then=_validate_link_response).href

    def get_upload_link_object(
        self,
        path: str,
        /,
        spoof_user_agent: bool = True,
        **kwargs
    ) -> ResourceUploadLinkObject:
        """
            Get a link to upload the file using the PUT request.
            This is similar to :any:`Client.get_upload_link()`, except it returns
            an instance of :any:`ResourceUploadLinkObject` which also contains
            an asynchronous operation ID.

            :param path: destination path
            :param overwrite: `bool`, determines whether to overwrite the destination
            :param fields: list of keys to be included in the response
            :param spoof_user_agent: `bool`, if `True` (default), the `User-Agent` header
                will be set to a special value, which should allow bypassing of
                Yandex.Disk's upload speed limit
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ParentNotFoundError: parent directory doesn't exist
            :raises PathExistsError: destination path already exists
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request
            :raises InsufficientStorageError: cannot upload file due to lack of storage space
            :raises UploadTrafficLimitExceededError: upload limit has been exceeded

            :returns: :any:`ResourceUploadLinkObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        # This is used to bypass Yandex.Disk's upload speed limit for some file types
        if spoof_user_agent:
            _add_spoof_user_agent_header(kwargs)

        return GetUploadLinkRequest(
            self.session, path, **kwargs
        ).send(yadisk=self)

    def _upload(self,
                get_upload_link_function: Callable,
                file_or_path: FileOrPath,
                dst_path: str, /, **kwargs) -> None:
        timeout = kwargs.get("timeout", ...)

        if timeout is ...:
            timeout = settings.DEFAULT_UPLOAD_TIMEOUT

        retry_interval = kwargs.get("retry_interval")

        if retry_interval is None:
            retry_interval = settings.DEFAULT_UPLOAD_RETRY_INTERVAL

        n_retries = kwargs.get("n_retries")

        if n_retries is None:
            n_retries = settings.DEFAULT_N_RETRIES

        # Number of retries for getting the upload link.
        # It is set to 0, unless the file is not seekable, in which case
        # we have to use a different retry scheme
        n_retries_for_upload_link = 0

        kwargs["timeout"] = timeout

        # Make sure we don't get any inconsistent behavior with header names
        kwargs["headers"] = CaseInsensitiveDict(kwargs.get("headers") or {})

        file: Any = None
        close_file = False
        file_position = 0
        iterator_factory = None

        session = self.session

        try:
            if isinstance(file_or_path, (str, bytes)):
                close_file = True
                file = self.open_file(file_or_path, "rb")
            elif callable(file_or_path):
                close_file = False
                iterator_factory = file_or_path
            else:
                close_file = False
                file = file_or_path

            if file is not None and file.seekable():
                file_position = file.tell()
            elif iterator_factory is None:
                n_retries, n_retries_for_upload_link = 0, n_retries

            def attempt() -> None:
                temp_kwargs = dict(kwargs)
                temp_kwargs["n_retries"] = n_retries_for_upload_link
                temp_kwargs["retry_interval"] = 0.0

                link = get_upload_link_function(dst_path, **temp_kwargs)

                # session.put() doesn't accept some of the passed parameters
                _filter_request_kwargs(temp_kwargs)

                temp_kwargs.setdefault("stream", True)

                # Disable keep-alive by default, since the upload server is random
                temp_kwargs["headers"].setdefault("Connection", "close")

                # This is generally not necessary, libraries like aiohttp
                # will generally always set this header while others might not
                # We're setting content-type here just to fix this inconsistency,
                # this makes testing easier
                temp_kwargs["headers"].setdefault("Content-Type", "application/octet-stream")

                if iterator_factory is not None:
                    payload = iterator_factory()
                elif file.seekable():
                    file.seek(file_position)
                    payload = file
                else:
                    # requests will try to seek the file to determine the payload size
                    # regardless of whether it is seekable() or not.
                    # To bypass this problem we pass the file as a generator instead.
                    payload = _read_file_as_generator(file)

                settings.logger.info(f"uploading file to {dst_path} at {link}")

                with session.send_request("PUT", link, data=payload, **temp_kwargs) as response:
                    if response.status != 201:
                        raise response.get_exception()

            auto_retry(attempt, n_retries, retry_interval)
        finally:
            if close_file and file is not None:
                file.close()

    def upload(
        self,
        file_or_path: FileOrPath,
        dst_path: str,
        /,
        **kwargs
    ) -> SyncResourceLinkObject:
        """
            Upload a file to disk.

            :param file_or_path: path, file-like object to be uploaded or
                                 a function that returns an iterator (or generator)
            :param dst_path: destination path
            :param overwrite: if `True`, the resource will be overwritten if it already exists,
                              an error will be raised otherwise
            :param spoof_user_agent: `bool`, if `True` (default), the `User-Agent` header
                will be set to a special value, which should allow bypassing of
                Yandex.Disk's upload speed limit
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ParentNotFoundError: parent directory doesn't exist
            :raises PathExistsError: destination path already exists
            :raises InsufficientStorageError: cannot upload file due to lack of storage space
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request
            :raises UploadTrafficLimitExceededError: upload limit has been exceeded

            :returns: :any:`SyncResourceLinkObject`, link to the destination resource
        """

        _apply_default_args(kwargs, self.default_args)

        self._upload(self.get_upload_link, file_or_path, dst_path, **kwargs)

        return SyncResourceLinkObject.from_path(dst_path, yadisk=self)

    def upload_by_link(self,
                       file_or_path: FileOrPath,
                       link: str, /, **kwargs) -> None:
        """
            Upload a file to disk using an upload link.

            :param file_or_path: path, file-like object to be uploaded or
                                 a function that returns an iterator (or generator)
            :param link: upload link
            :param overwrite: if `True`, the resource will be overwritten if it already exists,
                              an error will be raised otherwise
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises InsufficientStorageError: cannot upload file due to lack of storage space
        """

        _apply_default_args(kwargs, self.default_args)

        self._upload(lambda *args, **kwargs: link, file_or_path, "", **kwargs)

    def get_download_link(self, path: str, /, **kwargs) -> str:
        """
            Get a download link for a file (or a directory).

            :param path: path to the resource
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request

            :returns: `str`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return GetDownloadLinkRequest(
            self.session, path, fields=["href"], **kwargs
        ).send(yadisk=self, then=_validate_link_response).href

    def _download(
        self,
        get_download_link_function: Callable,
        src_path: str,
        file_or_path: FileOrPathDestination,
        /,
        **kwargs
    ) -> None:
        n_retries = kwargs.get("n_retries")

        if n_retries is None:
            n_retries = settings.DEFAULT_N_RETRIES

        # Number of retries for getting the download link.
        # It is set to 0, unless the file is not seekable, in which case
        # we have to use a different retry scheme
        n_retries_for_download_link = 0

        retry_interval = kwargs.get("retry_interval")

        if retry_interval is None:
            retry_interval = settings.DEFAULT_RETRY_INTERVAL

        timeout = kwargs.get("timeout", ...)

        if timeout is ...:
            timeout = settings.DEFAULT_TIMEOUT

        kwargs["timeout"] = timeout

        file: Any = None
        close_file = False
        file_position = 0

        session = self.session

        try:
            if isinstance(file_or_path, (str, bytes)):
                close_file = True
                file = self.open_file(file_or_path, "wb")
            else:
                close_file = False
                file = file_or_path

            if file.seekable():
                file_position = file.tell()
            else:
                n_retries, n_retries_for_download_link = 0, n_retries

            def attempt() -> None:
                temp_kwargs = dict(kwargs)
                temp_kwargs["n_retries"] = n_retries_for_download_link
                temp_kwargs["retry_interval"] = 0.0
                link = get_download_link_function(src_path, **temp_kwargs)

                # session.get() doesn't accept some of the passed parameters
                _filter_request_kwargs(temp_kwargs)

                temp_kwargs.setdefault("stream", True)

                if file.seekable():
                    file.seek(file_position)

                settings.logger.info(f"downloading file {src_path} from {link}")

                with session.send_request("GET", link, **temp_kwargs) as response:
                    # pycurl can't get status until the response is actually read
                    # in that case, status will be set to 0
                    if response.status == 0:
                        def consume(chunk: bytes) -> None:
                            if response.status not in (0, 200):
                                return

                            file.write(chunk)

                        response.download(consume)

                        if response.status != 200:
                            raise response.get_exception()
                    else:
                        if response.status != 200:
                            raise response.get_exception()

                        response.download(file.write)

            auto_retry(attempt, n_retries, retry_interval)
        finally:
            if close_file and file is not None:
                file.close()

    def download(
        self,
        src_path: str,
        file_or_path: FileOrPathDestination,
        /,
        **kwargs
    ) -> SyncResourceLinkObject:
        """
            Download the file.

            :param src_path: source path
            :param file_or_path: destination path or file-like object
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request

            :returns: :any:`SyncResourceLinkObject`, link to the source resource
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        self._download(self.get_download_link, src_path, file_or_path, **kwargs)

        return SyncResourceLinkObject.from_path(src_path, yadisk=self)

    def download_by_link(
        self,
        link: str,
        file_or_path: FileOrPathDestination,
        /,
        **kwargs
    ) -> None:
        """
            Download the file from the link.

            :param link: download link
            :param file_or_path: destination path or file-like object
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`
        """

        _apply_default_args(kwargs, self.default_args)

        self._download(lambda *args, **kwargs: link, "", file_or_path, **kwargs)

    def remove(self, path: str, /, **kwargs) -> Optional["SyncOperationLinkObject"]:
        """
            Remove the resource.

            :param path: path to the resource to be removed
            :param permanently: if `True`, the resource will be removed permanently,
                                otherwise, it will be just moved to the trash
            :param md5: `str`, MD5 hash of the file to remove
            :param force_async: forces the operation to be executed asynchronously
            :param wait: `bool`, if :code:`True`, the method will wait until the asynchronous operation is completed
            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises BadRequestError: MD5 check is only available for files
            :raises ResourceIsLockedError: resource is locked by another request
            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)

            :returns: :any:`SyncOperationLinkObject` if the operation is performed asynchronously, `None` otherwise
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return self._maybe_wait(DeleteRequest, path, **kwargs)

    def mkdir(self, path: str, /, **kwargs) -> SyncResourceLinkObject:
        """
            Create a new directory.

            :param path: path to the directory to be created
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ParentNotFoundError: parent directory doesn't exist
            :raises DirectoryExistsError: destination path already exists
            :raises InsufficientStorageError: cannot create directory due to lack of storage space
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request

            :returns: :any:`SyncResourceLinkObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return MkdirRequest(self.session, path, **kwargs).send(yadisk=self)

    def makedirs(self, path: str, /, **kwargs) -> SyncResourceLinkObject:
        """
            Create a new directory at `path`. If its parent directory doesn't
            exist it will also be created recursively.

            :param path: path to the directory to be created
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises DirectoryExistsError: destination path already exists
            :raises InsufficientStorageError: cannot create directory due to lack of storage space
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request

            :returns: :any:`SyncResourceLinkObject`
        """

        while True:
            try:
                return self.mkdir(path, **kwargs)
            except ParentNotFoundError as e:
                # We first have to remove the schema, otherwise posixpath.split()
                # may treat it as part of the path
                schema, path_without_schema = remove_path_schema(path)

                # Extract the parent directory
                head, tail = posixpath.split(path_without_schema)
                head = head.strip("/")

                if head == "":
                    # We should never find ourselves in this situation
                    raise e from None

                # Restore the schema
                if schema:
                    head = f"{schema}:/{head}"

                self.makedirs(head, **kwargs)

    def get_trash_meta(self, path: str, /, **kwargs) -> "SyncTrashResourceObject":
        """
            Get meta information about a trash resource.

            :param path: path to the trash resource
            :param limit: number of children resources to be included in the response
            :param offset: number of children resources to be skipped in the response
            :param preview_size: size of the file preview
            :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
            :param sort: `str`, field to be used as a key to sort children resources
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: :any:`SyncTrashResourceObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        # This is for internal error handling
        _then = kwargs.pop("_then", None)

        return GetTrashRequest(self.session, path, **kwargs).send(yadisk=self, then=_then)

    def trash_exists(self, path: str, /, **kwargs) -> bool:
        """
            Check whether the trash resource at `path` exists.

            :param path: path to the trash resource
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: `bool`
        """

        return _exists(self.get_trash_meta, path, **kwargs)

    def copy(
        self,
        src_path: str,
        dst_path: str,
        /,
        **kwargs
    ) -> Union[SyncResourceLinkObject, "SyncOperationLinkObject"]:
        """
            Copy `src_path` to `dst_path`.
            If the operation is performed asynchronously, returns the link to the operation,
            otherwise, returns the link to the newly created resource.

            :param src_path: source path
            :param dst_path: destination path
            :param overwrite: if `True` the destination path can be overwritten,
                              otherwise, an error will be raised
            :param force_async: forces the operation to be executed asynchronously
            :param fields: list of keys to be included in the response
            :param wait: `bool`, if :code:`True`, the method will wait until the asynchronous operation is completed
            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises PathExistsError: destination path already exists
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises InsufficientStorageError: cannot complete request due to lack of storage space
            :raises ResourceIsLockedError: resource is locked by another request
            :raises UploadTrafficLimitExceededError: upload limit has been exceeded
            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)

            :returns: :any:`SyncResourceLinkObject` or :any:`SyncOperationLinkObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return self._maybe_wait(CopyRequest, src_path, dst_path, **kwargs)

    def restore_trash(
        self,
        path: str,
        /,
        dst_path: Optional[str] = None,
        **kwargs
    ) -> Union[SyncResourceLinkObject, "SyncOperationLinkObject"]:
        """
            Restore a trash resource.
            Returns a link to the newly created resource or a link to the asynchronous operation.

            :param path: path to the trash resource to be restored
            :param dst_path: destination path
            :param overwrite: `bool`, determines whether the destination can be overwritten
            :param force_async: forces the operation to be executed asynchronously
            :param fields: list of keys to be included in the response
            :param wait: `bool`, if :code:`True`, the method will wait until the asynchronous operation is completed
            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises PathExistsError: destination path already exists
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request
            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)

            :returns: :any:`SyncResourceLinkObject` or :any:`SyncOperationLinkObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return self._maybe_wait(RestoreTrashRequest, path, dst_path=dst_path, **kwargs)

    def move(
        self,
        src_path: str,
        dst_path: str,
        /,
        **kwargs
    ) -> Union[SyncResourceLinkObject, "SyncOperationLinkObject"]:
        """
            Move `src_path` to `dst_path`.

            :param src_path: source path to be moved
            :param dst_path: destination path
            :param overwrite: `bool`, determines whether to overwrite the destination
            :param force_async: forces the operation to be executed asynchronously
            :param fields: list of keys to be included in the response
            :param wait: `bool`, if :code:`True`, the method will wait until the asynchronous operation is completed
            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises PathExistsError: destination path already exists
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request
            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)

            :returns: :any:`SyncResourceLinkObject` or :any:`SyncOperationLinkObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return self._maybe_wait(MoveRequest, src_path, dst_path, **kwargs)

    def rename(
        self,
        src_path: str,
        new_name: str,
        /,
        **kwargs
    ) -> Union[SyncResourceLinkObject, "SyncOperationLinkObject"]:
        """
            Rename `src_path` to have filename `new_name`.
            Does the same as `move()` but changes only the filename.

            :param src_path: source path to be moved
            :param new_name: target filename to rename to
            :param overwrite: `bool`, determines whether to overwrite the destination
            :param force_async: forces the operation to be executed asynchronously
            :param fields: list of keys to be included in the response
            :param wait: `bool`, if :code:`True`, the method will wait until the asynchronous operation is completed
            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises PathExistsError: destination path already exists
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request
            :raises ValueError: `new_name` is not a valid filename or `src_path` is root
            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)

            :returns: :any:`SyncResourceLinkObject` or :any:`SyncOperationLinkObject`
        """

        new_name = new_name.rstrip("/")

        if "/" in new_name or new_name in (".", "..", ""):
            raise ValueError(f"Invalid filename: {new_name}")

        # Remove schema first, otherwise PurePosixPath will treat it as part of the path
        schema, src_path_without_schema = remove_path_schema(src_path)
        sanitized_src_path = PurePosixPath(src_path_without_schema.strip("/"))

        if len(sanitized_src_path.parts) == 0:
            raise ValueError("Cannot rename root")

        dst_path = str(sanitized_src_path.parent / new_name)

        # Restore schema back
        if schema:
            dst_path = f"{schema}:/{dst_path}"

        return self.move(src_path, dst_path, **kwargs)

    def remove_trash(
        self,
        path: str,
        /,
        **kwargs
    ) -> Optional["SyncOperationLinkObject"]:
        """
            Remove a trash resource.

            :param path: path to the trash resource to be deleted
            :param force_async: forces the operation to be executed asynchronously
            :param fields: list of keys to be included in the response
            :param wait: `bool`, if :code:`True`, the method will wait until the asynchronous operation is completed
            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request
            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)

            :returns: :any:`SyncOperationLinkObject` if the operation is performed asynchronously, `None` otherwise
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return self._maybe_wait(DeleteTrashRequest, path, **kwargs)

    def publish(self, path: str, /, **kwargs) -> SyncResourceLinkObject:
        """
            Make a resource public.

            :param path: path to the resource to be published
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request

            :returns: :any:`SyncResourceLinkObject`, link to the resource
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return PublishRequest(self.session, path, **kwargs).send(yadisk=self)

    def unpublish(self, path: str, /, **kwargs) -> SyncResourceLinkObject:
        """
            Make a public resource private.

            :param path: path to the resource to be unpublished
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request

            :returns: :any:`SyncResourceLinkObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return UnpublishRequest(self.session, path, **kwargs).send(yadisk=self)

    def save_to_disk(
        self,
        public_key: str,
        /,
        **kwargs
    ) -> Union[SyncResourceLinkObject, "SyncOperationLinkObject"]:
        """
            Saves a public resource to the disk.
            Returns the link to the operation if it's performed asynchronously,
            or a link to the resource otherwise.

            :param public_key: public key or public URL of the public resource
            :param name: filename of the saved resource
            :param path: path to the copied resource in the public folder
            :param save_path: path to the destination directory (downloads directory by default)
            :param force_async: forces the operation to be executed asynchronously
            :param fields: list of keys to be included in the response
            :param wait: `bool`, if :code:`True`, the method will wait until the asynchronous operation is completed
            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request
            :raises InsufficientStorageError: cannot upload file due to lack of storage space
            :raises UploadTrafficLimitExceededError: upload limit has been exceeded
            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)

            :returns: :any:`SyncResourceLinkObject` or :any:`SyncOperationLinkObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return self._maybe_wait(SaveToDiskRequest, public_key, **kwargs)

    def get_public_meta(
        self,
        public_key: str,
        /,
        **kwargs
    ) -> "SyncPublicResourceObject":
        """
            Get meta-information about a public resource.

            :param public_key: public key or public URL of the public resource
            :param path: relative path to a resource in a public folder.
                         By specifying the key of the published folder in `public_key`,
                         you can request metainformation for any resource in the folder.
            :param offset: offset from the beginning of the list of nested resources
            :param limit: maximum number of nested elements to be included in the list
            :param sort: `str`, field to be used as a key to sort children resources
            :param preview_size: file preview size
            :param preview_crop: `bool`, allow preview crop
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: :any:`SyncPublicResourceObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        # This is for internal error handling
        _then = kwargs.pop("_then", None)

        return GetPublicMetaRequest(self.session, public_key, **kwargs).send(yadisk=self, then=_then)

    def public_exists(self, public_key: str, /, **kwargs) -> bool:
        """
            Check whether the public resource exists.

            :param public_key: public key or public URL of the public resource
            :param path: relative path to the resource within the public folder
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: `bool`
        """

        return _exists(self.get_public_meta, public_key, **kwargs)

    def public_listdir(
        self,
        public_key: str,
        /,
        **kwargs
    ) -> Generator["SyncPublicResourceObject", None, None]:
        """
            Get contents of a public directory.

            :param public_key: public key or public URL of the public resource
            :param path: relative path to the resource in the public folder.
                         By specifying the key of the published folder in `public_key`,
                         you can request contents of any nested folder.
            :param max_items: `int` or `None`, maximum number of returned items (`None` means unlimited)
            :param limit: number of children resources to be included in the response
            :param offset: number of children resources to be skipped in the response
            :param preview_size: size of the file preview
            :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises WrongResourceTypeError: resource is not a directory

            :returns: generator of :any:`SyncPublicResourceObject`
        """

        return _listdir(self.get_public_meta, public_key, **kwargs)

    def get_public_type(self, public_key: str, /, **kwargs) -> str:
        """
            Get public resource type.

            :param public_key: public key or public URL of the public resource
            :param path: relative path to the resource within the public folder
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: "file" or "dir"
        """

        return _get_type(self.get_public_meta, public_key, **kwargs)

    def is_public_dir(self, public_key: str, /, **kwargs) -> bool:
        """
            Check whether the public resource is a public directory.

            :param public_key: public key or public URL of the public resource
            :param path: relative path to the resource within the public folder
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: `True` if `public_key` is a directory, `False` otherwise (even if it doesn't exist)
        """

        try:
            return self.get_public_type(public_key, **kwargs) == "dir"
        except PathNotFoundError:
            return False

    def is_public_file(self, public_key: str, /, **kwargs) -> bool:
        """
            Check whether the public resource is a public file.

            :param public_key: public key or public URL of the public resource
            :param path: relative path to the resource within the public folder
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: `True` if `public_key` is a file, `False` otherwise (even if it doesn't exist)
        """

        try:
            return self.get_public_type(public_key, **kwargs) == "file"
        except PathNotFoundError:
            return False

    def trash_listdir(
        self,
        path: str,
        /,
        **kwargs
    ) -> Generator["SyncTrashResourceObject", None, None]:
        """
            Get contents of a trash resource.

            :param path: path to the directory in the trash bin
            :param max_items: `int` or `None`, maximum number of returned items (`None` means unlimited)
            :param limit: number of children resources to be included in the response
            :param offset: number of children resources to be skipped in the response
            :param preview_size: size of the file preview
            :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises WrongResourceTypeError: resource is not a directory

            :returns: generator of :any:`SyncTrashResourceObject`
        """

        return _listdir(self.get_trash_meta, path, **kwargs)

    def get_trash_type(self, path: str, /, **kwargs) -> str:
        """
            Get trash resource type.

            :param path: path to the trash resource
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: "file" or "dir"
        """

        return _get_type(self.get_trash_meta, path, **kwargs)

    def is_trash_dir(self, path: str, /, **kwargs) -> bool:
        """
            Check whether `path` is a trash directory.

            :param path: path to the trash resource
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: `True` if `path` is a directory, `False` otherwise (even if it doesn't exist)
        """

        try:
            return self.get_trash_type(path, **kwargs) == "dir"
        except PathNotFoundError:
            return False

    def is_trash_file(self, path: str, /, **kwargs) -> bool:
        """
            Check whether `path` is a trash file.

            :param path: path to the trash resource
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: `True` if `path` is a file, `False` otherwise (even if it doesn't exist)
        """

        try:
            return self.get_trash_type(path, **kwargs) == "file"
        except PathNotFoundError:
            return False

    def get_public_resources(self, **kwargs) -> "SyncPublicResourcesListObject":
        """
            Get a list of public resources.

            :param offset: offset from the beginning of the list
            :param limit: maximum number of elements in the list
            :param preview_size: size of the file preview
            :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
            :param type: filter based on type of resources ("file" or "dir")
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: :any:`SyncPublicResourcesListObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return GetPublicResourcesRequest(self.session, **kwargs).send(yadisk=self)

    def get_all_public_resources(
        self,
        *,
        max_items: Optional[int] = None,
        **kwargs
    ) -> Generator[SyncPublicResourceObject, None, None]:
        """
            Get a list of all public resources.

            :param max_items: `int` or `None`, maximum number of returned items (`None` means unlimited)
            :param offset: offset from the beginning of the list
            :param limit: maximum number of elements in the list
            :param preview_size: size of the file preview
            :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
            :param type: filter based on type of resources ("file" or "dir")
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: generator of :any:`SyncPublicResourceObject`
        """

        if kwargs.get("offset") is None:
            kwargs["offset"] = 0

        if kwargs.get("limit") is None:
            kwargs["limit"] = 100

        remaining_items = max_items

        while True:
            # Do not query more items than necessary
            if remaining_items is not None:
                kwargs["limit"] = min(remaining_items, kwargs["limit"])

            files = self.get_public_resources(**kwargs).items or []
            file_count = len(files)

            yield from files[:remaining_items]

            if remaining_items is not None:
                remaining_items -= file_count

                if remaining_items <= 0:
                    break

            if file_count < kwargs["limit"]:
                break

            kwargs["offset"] += kwargs["limit"]

    def patch(self,
              path: str,
              properties: dict, /, **kwargs) -> "SyncResourceObject":
        """
            Update custom properties of a resource.

            :param path: path to the resource
            :param properties: `dict`, custom properties to update
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request

            :returns: :any:`ResourceObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return PatchRequest(self.session, path, properties, **kwargs).send(yadisk=self)

    def _get_files_some(self, **kwargs) -> List["SyncResourceObject"]:
        response: "SyncFilesResourceListObject" = FilesRequest(
            self.session, **kwargs).send(yadisk=self)

        if response.items is None:
            raise InvalidResponseError("Response did not contain key field")

        return response.items

    def get_files(
        self,
        *,
        max_items: Optional[int] = None,
        **kwargs
    ) -> Generator["SyncResourceObject", None, None]:
        """
            Get a flat list of all files (that doesn't include directories).

            :param offset: offset from the beginning of the list
            :param max_items: `int` or `None`, maximum number of returned items (`None` means unlimited)
            :param limit: number of list elements to be included in each response
            :param media_type: type of files to include in the list
            :param sort: `str`, field to be used as a key to sort children resources
            :param preview_size: size of the file preview
            :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: generator of :any:`ResourceObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        if kwargs.get("offset") is None:
            kwargs["offset"] = 0

        if kwargs.get("limit") is None:
            kwargs["limit"] = 200

        remaining_items = max_items

        while True:
            # Do not query more items than necessary
            if remaining_items is not None:
                kwargs["limit"] = min(remaining_items, kwargs["limit"])

            files = self._get_files_some(**kwargs)
            file_count = len(files)

            yield from files[:remaining_items]

            if remaining_items is not None:
                remaining_items -= file_count

                if remaining_items <= 0:
                    break

            if file_count < kwargs["limit"]:
                break

            kwargs["offset"] += kwargs["limit"]

    def get_last_uploaded(self, **kwargs) -> List["SyncResourceObject"]:
        """
            Get the list of latest uploaded files sorted by upload date.

            :param limit: maximum number of elements in the list
            :param media_type: type of files to include in the list
            :param preview_size: size of the file preview
            :param preview_crop: `bool`, cut the preview to the size specified in the `preview_size`
            :param fields: list of keys to be included in the response
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ForbiddenError: application doesn't have enough rights for this request

            :returns: generator of :any:`ResourceObject`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        response: "SyncLastUploadedResourceListObject" = LastUploadedRequest(
            self.session, **kwargs
        ).send(yadisk=self)

        if response.items is None:
            raise InvalidResponseError("Response did not contain key field")

        return response.items

    def upload_url(
        self,
        url: str,
        path: str,
        /,
        **kwargs
    ) -> "SyncOperationLinkObject":
        """
            Upload a file from URL.

            :param url: source URL
            :param path: destination path
            :param disable_redirects: `bool`, forbid redirects
            :param fields: list of keys to be included in the response
            :param wait: `bool`, if :code:`True`, the method will wait until the asynchronous operation is completed
            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises ParentNotFoundError: parent directory doesn't exist
            :raises PathExistsError: destination path already exists
            :raises InsufficientStorageError: cannot upload file due to lack of storage space
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request
            :raises UploadTrafficLimitExceededError: upload limit has been exceeded
            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)

            :returns: :any:`SyncOperationLinkObject`, link to the asynchronous operation
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return self._maybe_wait(UploadURLRequest, url, path, **kwargs)

    def get_public_download_link(self, public_key: str, /, **kwargs) -> str:
        """
            Get a download link for a public resource.

            :param public_key: public key or public URL of the public resource
            :param path: relative path to the resource within the public folder
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request

            :returns: `str`
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return GetPublicDownloadLinkRequest(
            self.session, public_key, fields=["href"], **kwargs
        ).send(yadisk=self, then=_validate_link_response).href

    def download_public(
        self,
        public_key: str,
        file_or_path: FileOrPathDestination,
        /,
        **kwargs
    ) -> SyncPublicResourceLinkObject:
        """
            Download the public resource.

            :param public_key: public key or public URL of the public resource
            :param file_or_path: destination path or file-like object
            :param path: relative path to the resource within the public folder
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises PathNotFoundError: resource was not found on Disk
            :raises ForbiddenError: application doesn't have enough rights for this request
            :raises ResourceIsLockedError: resource is locked by another request

            :returns: :any:`SyncPublicResourceLinkObject`
        """

        _apply_default_args(kwargs, self.default_args)

        self._download(
            lambda public_key, **kwargs: self.get_public_download_link(public_key, **kwargs),
            public_key, file_or_path, **kwargs)

        return SyncPublicResourceLinkObject.from_public_key(public_key, yadisk=self)

    def get_operation_status(self, operation_id: str, /, **kwargs) -> OperationStatus:
        """
            Get operation status.

            :param operation_id: ID of the operation or a link
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises OperationNotFoundError: requested operation was not found

            :returns: `str`, :code:`"in-progress"` indicates that the operation
                      is currently running, :code:`"success"` indicates that
                      the operation was successful, :code:`"failed"` means that
                      the operation failed
        """

        _apply_default_args(kwargs, self.default_args)
        _add_authorization_header(kwargs, self.token)

        return GetOperationStatusRequest(
            self.session, operation_id, fields=["status"], **kwargs
        ).send(yadisk=self).status

    def wait_for_operation(
        self,
        operation_id: str,
        /,
        *,
        poll_interval: float = 1.0,
        poll_timeout: Optional[float] = None,
        **kwargs
    ) -> None:
        """
            Wait until an operation is completed. If the operation fails, an
            exception is raised. Waiting is performed by calling :any:`time.sleep`.

            :param operation_id: ID of the operation or a link
            :param poll_interval: `float`, interval in seconds between subsequent operation status queries
            :param poll_timeout: `float` or `None`, total polling timeout (`None` means no timeout),
                                 if this timeout is exceeded, an exception is raised
            :param timeout: `float` or `tuple`, request timeout
            :param headers: `dict` or `None`, additional request headers
            :param n_retries: `int`, maximum number of retries
            :param retry_interval: delay between retries in seconds
            :param retry_on: `tuple`, additional exception classes to retry on
            :param requests_args: `dict`, additional parameters for :any:`RequestsSession`
            :param httpx_args: `dict`, additional parameters for :any:`HTTPXSession`
            :param curl_options: `dict`, additional options for :any:`PycURLSession`
            :param kwargs: any other parameters, accepted by :any:`Session.send_request()`

            :raises OperationNotFoundError: requested operation was not found
            :raises AsyncOperationFailedError: requested operation failed
            :raises AsyncOperationPollingTimeoutError: requested operation did not
                                                       complete in specified time
                                                       (when `poll_timeout` is not `None`)
        """

        poll_start_time = time.monotonic()

        while (status := self.get_operation_status(operation_id, **kwargs)) == "in-progress":
            if poll_timeout is not None:
                total_poll_duration = time.monotonic() - poll_start_time

                if total_poll_duration >= poll_timeout:
                    raise AsyncOperationPollingTimeoutError("Asynchronous operation did not complete in specified time")

            time.sleep(poll_interval)

        if status != "success":
            raise AsyncOperationFailedError("Asynchronous operation failed")



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from ..exceptions import (
    RequestError, TooManyRedirectsError,
    RequestTimeoutError, YaDiskConnectionError
)

from .._session import Session, Response
from ..utils import CaseInsensitiveDict
from .._typing_compat import Dict
from ..types import JSON, ConsumeCallback, HTTPMethod, Headers, Payload

from typing import Any, Optional, Union

import threading

import requests

__all__ = ["RequestsSession"]


def convert_requests_exception(exc: requests.RequestException) -> Union[RequestError, requests.RequestException]:
    if isinstance(exc, requests.exceptions.TooManyRedirects):
        return TooManyRedirectsError(str(exc))
    elif isinstance(exc, requests.exceptions.Timeout):
        return RequestTimeoutError(str(exc))
    elif isinstance(exc, requests.exceptions.ConnectionError):
        return YaDiskConnectionError(str(exc))
    elif isinstance(exc, requests.RequestException):
        return RequestError(str(exc))
    else:
        return exc


class RequestsResponse(Response):
    def __init__(self, response: requests.Response):
        super().__init__()

        self._response = response
        self.status = self._response.status_code

    def json(self) -> JSON:
        try:
            return self._response.json()
        except RuntimeError as e:
            raise ValueError(f"Could not parse JSON: {e}") from e

    def download(self, consume_callback: ConsumeCallback) -> None:
        try:
            for chunk in self._response.iter_content(8192):
                consume_callback(chunk)
        except requests.RequestException as e:
            raise convert_requests_exception(e) from e

    def close(self) -> None:
        self._response.close()


class RequestsSession(Session):
    """
        .. _requests: https://pypi.org/project/requests

        :any:`Session` implementation using the `requests`_ library.

        All arguments passed in the constructor are directly forwared to :any:`requests.Session`.

        :ivar requests_session: underlying instance of :any:`requests.Session`

        .. note::
           Internally, this class creates thread-local instances of
           :any:`requests.Session`, since it is not currently guaranteed to be
           thread safe.
           Calling :any:`Session.close()` will close all thread-local sessions
           managed by this object.

        To pass `requests`-specific arguments from :any:`Client` use :code:`requests_args` keyword argument.

        Usage example:

        .. code:: python

           import yadisk

           with yadisk.Client(..., session="requests") as client:
               client.get_meta(
                   "/my_file.txt",
                   n_retries=5,
                   requests_args={
                       "proxies": {"https": "http://example.com:1234"},
                       "verify": False
                   }
                )
    """

    def __init__(self, *args, **kwargs):
        self._args, self._kwargs = args, kwargs
        self._local = threading.local()
        self._sessions = []

    @property
    def requests_session(self) -> requests.Session:
        if not hasattr(self._local, "session"):
            self._local.session = requests.Session(*self._args, **self._kwargs)
            self._sessions.append(self._local.session)

        return self._local.session

    def _close_local(self) -> None:
        if not hasattr(self._local, "session"):
            return

        session = self._local.session

        session.close()
        self._sessions.remove(session)

    def send_request(
        self,
        method: HTTPMethod,
        url: str,
        *,
        params: Optional[Dict[str, Any]] = None,
        data: Optional[Payload] = None,
        headers: Optional[Headers] = None,
        stream: bool = False,
        **kwargs
    ) -> Response:
        requests_headers = CaseInsensitiveDict(self.requests_session.headers)

        requests_headers.update(headers or {})

        converted_kwargs: Dict[str, Any] = {
            "params": params,
            "data": data,
            "headers": requests_headers,
            "stream": stream
        }

        if "timeout" in kwargs:
            converted_kwargs["timeout"] = kwargs["timeout"]

        if "requests_args" in kwargs:
            converted_kwargs.update(kwargs["requests_args"] or {})

        try:
            return RequestsResponse(
                self.requests_session.request(method, url, **converted_kwargs)
            )
        except requests.exceptions.RequestException as e:
            raise convert_requests_exception(e) from e

    def close(self) -> None:
        while self._sessions:
            session = self._sessions.pop()
            session.close()



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from typing import TYPE_CHECKING, Any, Optional, TypeVar
from ._typing_compat import Dict
from .types import (
    AsyncConsumeCallback, JSON, HTTPMethod, AsyncPayload, Headers, TimeoutParameter
)
from .objects import ErrorObject
from .utils import get_exception

if TYPE_CHECKING:  # pragma: no cover
    from .exceptions import YaDiskError

__all__ = ["AsyncResponse", "AsyncSession"]


class AsyncResponse:
    """
        Represents an HTTP response.

        In case an error occurs, methods of this class should throw one of exceptions
        derived from :any:`YaDiskError`.

        :ivar status: `int`, HTTP status code
    """

    _Self = TypeVar("_Self", bound="AsyncResponse")

    status: int

    def __init__(self) -> None:
        """Constructs an :any:`AsyncResponse` object."""

        self.status = 0

    async def json(self) -> JSON:
        """
            Returns JSON-content of the response (parses JSON).

            .. note::
               This is an abstract method that needs to be implemented.

            :raises RequestError: could not receive the response's body
            :raises ValueError: could not parse JSON

            :returns: `dict`, `list`, `str`, `int`, `float` or `None`
        """
        raise NotImplementedError

    async def download(self, consume_callback: AsyncConsumeCallback) -> None:
        """
            Downloads response's content.

            .. note::
               This is an abstract method that needs to be implemented.

            :param consume_callback: regular or async function, takes one
                                     parameter - chunk of data (bytes),
                                     consumes the chunk (e.g. by writing to a file)

            :raises RequestError: could not receive the response's body
        """
        raise NotImplementedError

    async def get_exception(self) -> "YaDiskError":
        """
            Convenience wrapper for :any:`yadisk.utils.get_exception`.

            :returns: :any:`YaDiskError`
        """

        try:
            js = await self.json()
        except ValueError:
            js = None

        error = ErrorObject(js)

        return get_exception(self, error)

    async def close(self) -> None:
        """
            Closes the response and releases the underlying connection into the pool.

            .. note::
               This is an abstract method that needs to be implemented.
        """
        raise NotImplementedError

    async def __aenter__(self: _Self) -> _Self:
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        """Closes the response and releases the underlying connection into the pool"""

        await self.close()


class AsyncSession:
    """
        HTTP session class. Maintains open connections, stores headers and
        some other request parameters.

        Must be explicitly closed (can be done using the `with` statement).
    """

    _Self = TypeVar("_Self", bound="AsyncSession")

    async def send_request(self,
                           method: HTTPMethod,
                           url: str,
                           *,
                           params:  Optional[Dict[str, Any]] = None,
                           data:    Optional[AsyncPayload] = None,
                           timeout: TimeoutParameter = ...,
                           headers: Optional[Headers] = None,
                           stream:  bool = False,
                           **kwargs) -> AsyncResponse:
        """
            Sends an HTTP request with given parameters.
            In case an error occurs, the method should throw one of exceptions
            derived from :any:`YaDiskError`.
            Additional keyword arguments may be passed, they may be forwarded
            to the underlying HTTP client without modification.

            .. note::
               This is an abstract method that needs to be implemented.

            :param method: `str`, HTTP method
            :param url: `str`, URL
            :param params: `dict`, GET parameters
            :param data: `bytes`, iterator (possibly async) or a file-like object (possible async),
                         data to be sent in the request body
            :param headers: `dict`, additional headers to be set
            :param timeout: request timeout, a `tuple` of `(read timeout, connect timeout)`,
                            `float` or `None` (no timeout)
            :param stream: `bool`, if `False`, the response content will be immediately downloaded

            :returns: :any:`Response`, response object
        """
        raise NotImplementedError

    async def close(self) -> None:
        """
            Closes the session.

            .. note::
               This is an abstract method that needs to be implemented.
        """
        raise NotImplementedError

    async def __aenter__(self: _Self) -> _Self:
        return self

    async def __aexit__(self, *args, **kwargs) -> None:
        """Closes the session."""
        return await self.close()



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from .api_request import APIRequest
from ..objects import DiskInfoObject
from ..exceptions import InvalidResponseError

from typing import Optional, TYPE_CHECKING
from .._typing_compat import Iterable

if TYPE_CHECKING:  # pragma: no cover
    from ..types import AnySession, JSON

__all__ = ["DiskInfoRequest"]


class DiskInfoRequest(APIRequest):
    """
        A request to get disk information.

        :param session: an instance of :any:`Session` or :any:`AsyncSession` with prepared headers
        :param fields: list of keys to be included in the response

        :returns: :any:`DiskInfoObject`
    """

    method = "GET"
    path = "/v1/disk"

    def __init__(
        self,
        session: "AnySession",
        fields: Optional[Iterable[str]] = None,
        **kwargs
    ) -> None:
        APIRequest.__init__(self, session, **kwargs)

        if fields is not None:
            self.params["fields"] = ",".join(fields)

    def process_json(self, js: "JSON", **kwargs) -> DiskInfoObject:
        if not isinstance(js, dict):
            raise InvalidResponseError("Yandex.Disk returned invalid JSON")

        return DiskInfoObject(js)



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

from ._yadisk_object import *
from ._error_object import *
from ._link_object import *
from ._disk import *
from ._resources import *
from ._operations import *
from ._auth import *

__all__ = [  # noqa: RUF022
    # _yadisk_object
    "YaDiskObject",

    # _error_object
    "ErrorObject",

    # _link_object
    "LinkObject",

    # _disk
    "DiskInfoObject",
    "SystemFoldersObject",
    "UserObject",
    "UserPublicInfoObject",

    # _resources
    "CommentIDsObject",
    "EXIFObject",
    "FilesResourceListObject",
    "SyncFilesResourceListObject",
    "AsyncFilesResourceListObject",
    "LastUploadedResourceListObject",
    "SyncLastUploadedResourceListObject",
    "AsyncLastUploadedResourceListObject",
    "PublicResourcesListObject",
    "SyncPublicResourcesListObject",
    "AsyncPublicResourcesListObject",
    "ResourceListObject",
    "SyncResourceListObject",
    "AsyncResourceListObject",
    "ResourceObject",
    "SyncResourceObject",
    "AsyncResourceObject",
    "ResourceUploadLinkObject",
    "ShareInfoObject",
    "PublicResourceObject",
    "SyncPublicResourceObject",
    "AsyncPublicResourceObject",
    "PublicResourceListObject",
    "SyncPublicResourceListObject",
    "AsyncPublicResourceListObject",
    "TrashResourceObject",
    "SyncTrashResourceObject",
    "AsyncTrashResourceObject",
    "TrashResourceListObject",
    "SyncTrashResourceListObject",
    "AsyncTrashResourceListObject",
    "ResourceLinkObject",
    "SyncResourceLinkObject",
    "AsyncResourceLinkObject",
    "PublicResourceLinkObject",
    "SyncPublicResourceLinkObject",
    "AsyncPublicResourceLinkObject",
    "ResourceDownloadLinkObject",

    # _operations
    "OperationStatusObject",
    "OperationLinkObject",
    "SyncOperationLinkObject",
    "AsyncOperationLinkObject",

    # _auth
    "TokenObject",
    "TokenRevokeStatusObject",
    "DeviceCodeObject"
]



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.

import asyncio
from collections import defaultdict
import sys
import time

from .objects import ErrorObject
from .exceptions import *
from . import settings

from typing import Any, Optional, Union, TypeVar

from ._typing_compat import Callable, Awaitable, Dict, Tuple, Type
from .types import AnyResponse

__all__ = ["CaseInsensitiveDict", "async_auto_retry", "auto_retry", "get_exception"]


class _UnexpectedRequestError(YaDiskError):
    # Used for testing (see tests/disk_gateway.py)
    pass


if sys.version_info >= (3, 11) and hasattr(Exception, "add_note"):
    def _add_exception_note(exc: Exception, note: str) -> None:
        exc.add_note(note)
else:
    def _add_exception_note(exc: Exception, note: str) -> None:
        pass


EXCEPTION_MAP: Dict[int, Dict[str, Type[YaDiskError]]] = {
    400: defaultdict(
        lambda: BadRequestError,
        {
            "FieldValidationError":   FieldValidationError,
            "authorization_pending":  AuthorizationPendingError,
            "invalid_client":         InvalidClientError,
            "invalid_grant":          InvalidGrantError,
            "bad_verification_code":  BadVerificationCodeError,
            "unsupported_token_type": UnsupportedTokenTypeError
        }
    ),
    401: defaultdict(lambda: UnauthorizedError),
    403: defaultdict(lambda: ForbiddenError),
    404: defaultdict(
        lambda: NotFoundError,
        {
            "DiskNotFoundError":          PathNotFoundError,
            "DiskOperationNotFoundError": OperationNotFoundError
        }
    ),
    406: defaultdict(lambda: NotAcceptableError),
    409: defaultdict(
        lambda: ConflictError,
        {
            "DiskPathDoesntExistsError":              ParentNotFoundError,
            "DiskPathPointsToExistentDirectoryError": DirectoryExistsError,
            "DiskResourceAlreadyExistsError":         PathExistsError,
            "MD5DifferError":                         MD5DifferError
        }
    ),
    410: defaultdict(lambda: GoneError),
    413: defaultdict(lambda: PayloadTooLargeError),
    415: defaultdict(lambda: UnsupportedMediaError),
    423: defaultdict(
        lambda: LockedError,
        {
            "DiskResourceLockedError":        ResourceIsLockedError,
            "DiskUploadTrafficLimitExceeded": UploadTrafficLimitExceededError
        }
    ),
    429: defaultdict(
        lambda: TooManyRequestsError,
        {"DiskResourceDownloadLimitExceededError": ResourceDownloadLimitExceededError}
    ),
    500: defaultdict(lambda: InternalServerError),
    502: defaultdict(lambda: BadGatewayError),
    503: defaultdict(lambda: UnavailableError),
    504: defaultdict(lambda: GatewayTimeoutError),
    507: defaultdict(lambda: InsufficientStorageError),

    # This is a special value for testing
    499: defaultdict(lambda: _UnexpectedRequestError)
}


def get_exception(response: AnyResponse, error: Optional[ErrorObject]) -> YaDiskError:
    """
        Get an exception instance based on response, assuming the request has failed.

        :param response: an instance of :any:`Response` or :any:`AsyncResponse`
        :param error: an instance of :any:`ErrorObject` or `None`

        :returns: an exception instance, subclass of :any:`YaDiskError`
    """

    exc_group = EXCEPTION_MAP.get(response.status, None)

    if exc_group is None:
        return UnknownYaDiskError(f"Unknown Yandex.Disk error: status code {response.status}")

    if error is not None:
        msg = error.message or ""
        desc = error.description or ""
        error_name = error.error or ""
    else:
        msg = ""
        desc = ""
        error_name = ""

    exc = exc_group[error_name]

    exc_message = ""

    if msg:
        exc_message = msg

    if desc:
        if exc_message:
            exc_message += " | "

        exc_message += f"Error description: {desc.rstrip('.')}. Error code: {error_name}"

    if error_name:
        if exc_message:
            exc_message += " | "

        exc_message += f"Error code: {error_name}"

    if exc_message:
        exc_message += " | "

    exc_message += f"Status code: {response.status}"

    return exc(error_name, exc_message, response)


T = TypeVar("T")


def auto_retry(
    func: Callable[..., T],
    n_retries: Optional[int] = None,
    retry_interval: Optional[Union[int, float]] = None,
    args: Optional[Tuple] = None,
    kwargs: Optional[Dict[str, Any]] = None,
    retry_on: Tuple[Type[Exception], ...] = tuple()
) -> T:
    """
        Attempt to perform a request with automatic retries.
        A retry is triggered by :any:`RequestError` or :any:`RetriableYaDiskError`,
        unless the raised exception has :code:`disable_retry` set to :code:`True`.

        :param func: function to run, must not require any arguments
        :param n_retries: `int`, maximum number of retries
        :param retry_interval: `int` or `float`, delay between retries (in seconds)
        :param args: `tuple` or `None`, additional arguments for `func`
        :param kwargs: `dict` or `None`, additional keyword arguments for `func`
        :param retry_on: `tuple`, additional exception classes to retry on

        :returns: return value of func()
    """

    if n_retries is None:
        n_retries = settings.DEFAULT_N_RETRIES

    if retry_interval is None:
        retry_interval = settings.DEFAULT_RETRY_INTERVAL

    if args is None:
        args = tuple()

    if kwargs is None:
        kwargs = {}

    exceptions: Tuple[Type[Exception], ...] = (RequestError, RetriableYaDiskError, *retry_on)

    for i in range(n_retries + 1):
        try:
            return func(*args, **kwargs)
        except exceptions as e:
            if i == n_retries or (isinstance(e, YaDiskError) and e.disable_retry):
                settings.logger.info(
                    f"not triggering an automatic retry: ({i + 1} out of {n_retries}), got {e.__class__.__name__}: {e}"
                )

                if i:
                    _add_exception_note(e, f"Got the error after {i} retry attempts")

                raise

            settings.logger.info(
                f"automatic retry triggered: ({i + 1} out of {n_retries}), got {e.__class__.__name__}: {e}"
            )

        if retry_interval:
            time.sleep(retry_interval)

    # This should never be reachable
    raise AssertionError()


async def async_auto_retry(
    func: Union[Callable[..., Any], Callable[..., Awaitable[Any]]],
    n_retries: Optional[int] = None,
    retry_interval: Optional[Union[int, float]] = None,
    args: Optional[Tuple] = None,
    kwargs: Optional[Dict[str, Any]] = None,
    retry_on: Tuple[Type[Exception], ...] = tuple()
) -> Any:
    """
        Attempt to perform a request with automatic retries.
        A retry is triggered by :any:`RequestError` or :any:`RetriableYaDiskError`,
        unless the raised exception has :code:`disable_retry` set to :code:`True`.

        :param func: function to run, must not require any arguments
        :param n_retries: `int`, maximum number of retries
        :param retry_interval: `int` or `float`, delay between retries (in seconds)
        :param args: `tuple` or `None`, additional arguments for `func`
        :param kwargs: `dict` or `None`, additional keyword arguments for `func`
        :param retry_on: `tuple`, additional exception classes to retry on

        :returns: return value of func()
    """

    if n_retries is None:
        n_retries = settings.DEFAULT_N_RETRIES

    if retry_interval is None:
        retry_interval = settings.DEFAULT_RETRY_INTERVAL

    if args is None:
        args = tuple()

    if kwargs is None:
        kwargs = {}

    is_coro = asyncio.iscoroutinefunction(func)

    # Suppress false type hint errors
    callback: Any = func

    exceptions: Tuple[Type[Exception], ...] = (RequestError, RetriableYaDiskError, *retry_on)

    for i in range(n_retries + 1):
        try:
            if is_coro:
                return await callback(*args, **kwargs)
            else:
                return callback(*args, **kwargs)
        except exceptions as e:
            if i == n_retries or (isinstance(e, YaDiskError) and e.disable_retry):
                settings.logger.info(
                    f"not triggering an automatic retry: ({i + 1} out of {n_retries}), got {e.__class__.__name__}: {e}"
                )

                if i:
                    _add_exception_note(e, f"Got the error after {i} retry attempts")

                raise

            settings.logger.info(
                f"automatic retry triggered: ({i + 1} out of {n_retries}), got {e.__class__.__name__}: {e}"
            )

        if retry_interval:
            await asyncio.sleep(retry_interval)

    # This should never be reachable
    raise AssertionError()


class CaseInsensitiveDict(dict):
    """A case-insensitive dictionary. All keys are converted to lowercase."""

    @classmethod
    def _k(cls, key: str) -> str:
        return key.lower()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._convert_keys()

    def __getitem__(self, key: str) -> Any:
        return super().__getitem__(self.__class__._k(key))

    def __setitem__(self, key: str, value: Any) -> None:
        super().__setitem__(self.__class__._k(key), value)

    def __delitem__(self, key: str) -> Any:
        return super().__delitem__(self.__class__._k(key))

    def __contains__(self, key: Any) -> bool:
        return super().__contains__(self.__class__._k(key))

    def pop(self, key: str, /, *args, **kwargs) -> Any:
        return super().pop(self.__class__._k(key), *args, **kwargs)

    def get(self, key: str, /, *args, **kwargs) -> Any:
        return super().get(self.__class__._k(key), *args, **kwargs)

    def setdefault(self, key: str, *args, **kwargs) -> Any:
        return super().setdefault(self.__class__._k(key), *args, **kwargs)

    def update(self, *args, **kwargs) -> None:
        super().update(*(self.__class__(arg) for arg in args), **self.__class__(kwargs))

    def _convert_keys(self) -> None:
        for k in list(self.keys()):
            v = super(CaseInsensitiveDict, self).pop(k)
            self.__setitem__(k, v)



# -*- coding: utf-8 -*-
# Copyright © 2024 Ivan Konovalov

# This file is part of a Python library yadisk.

# This library is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.



# -*- coding: utf-8 -*-

import yadisk
from yadisk.types import HTTPMethod

from urllib.parse import urljoin, urlparse

from typing import Optional

__all__ = ["AsyncTestSession", "TestSession"]

def _ensure_trailing_slash(s: str) -> str:
    if not s.endswith("/"):
        return s + "/"

    return s

class TestSession(yadisk.Session):
    __test__ = False

    def __init__(
        self,
        session:           yadisk.Session,
        disk_base_url:     Optional[str] = None,
        auth_base_url:     Optional[str] = None,
        download_base_url: Optional[str] = None,
        upload_base_url:   Optional[str] = None
    ):
        self._session          = session
        self.disk_base_url     = disk_base_url
        self.auth_base_url     = auth_base_url
        self.download_base_url = download_base_url
        self.upload_base_url   = upload_base_url

    def send_request(self, method: HTTPMethod, url: str, **kwargs) -> yadisk.Response:
        url_parsed = urlparse(url)
        kwargs.setdefault("headers", {})

        new_base_url: Optional[str] = None

        if url_parsed.hostname == "cloud-api.yandex.net":
            new_base_url = self.disk_base_url
        elif url_parsed.hostname == "oauth.yandex.ru":
            new_base_url = self.auth_base_url
        elif url_parsed.hostname == "downloader.disk.yandex.ru":
            new_base_url = self.download_base_url
        else:
            subdomain, _, domain = (url_parsed.hostname or "").partition(".")

            assert domain == "disk.yandex.net", f"Got unexpected URL: {url}"
            assert subdomain.startswith("uploader"), f"Got unexpected URL: {url}"

            if self.upload_base_url is not None:
                new_base_url = f"{self.upload_base_url.rstrip('/')}/{subdomain}"

        if new_base_url is not None:
            url = urljoin(_ensure_trailing_slash(new_base_url), url_parsed.path.lstrip("/"))
            if url_parsed.query:
                url += "?" + url_parsed.query

        return self._session.send_request(method, url, **kwargs)

    def close(self) -> None:
        self._session.close()

class AsyncTestSession(yadisk.AsyncSession):
    __test__ = False

    def __init__(
        self,
        session:           yadisk.AsyncSession,
        disk_base_url:     Optional[str] = None,
        auth_base_url:     Optional[str] = None,
        download_base_url: Optional[str] = None,
        upload_base_url:   Optional[str] = None,
    ):
        self._session          = session
        self.disk_base_url     = disk_base_url
        self.auth_base_url     = auth_base_url
        self.download_base_url = download_base_url
        self.upload_base_url   = upload_base_url

    async def send_request(self, method: HTTPMethod, url: str, **kwargs) -> yadisk.AsyncResponse:
        url_parsed = urlparse(url)
        kwargs.setdefault("headers", {})

        new_base_url: Optional[str] = None

        if url_parsed.hostname == "cloud-api.yandex.net":
            new_base_url = self.disk_base_url
        elif url_parsed.hostname == "oauth.yandex.ru":
            new_base_url = self.auth_base_url
        elif url_parsed.hostname == "downloader.disk.yandex.ru":
            new_base_url = self.download_base_url
        else:
            subdomain, _, domain = (url_parsed.hostname or "").partition(".")

            assert domain == "disk.yandex.net", f"Got unexpected URL: {url}"
            assert subdomain.startswith("uploader"), f"Got unexpected URL: {url}"

            if self.upload_base_url is not None:
                new_base_url = f"{self.upload_base_url.rstrip('/')}/{subdomain}"

        if new_base_url is not None:
            url = urljoin(_ensure_trailing_slash(new_base_url), url_parsed.path.lstrip("/"))
            if url_parsed.query:
                url += "?" + url_parsed.query

        return await self._session.send_request(method, url, **kwargs)

    async def close(self) -> None:
        await self._session.close()
